@@ -3,8 +3,11 @@ import os.path
 import json
 import traceback
 import glob
+import re
 from typing import Union, Optional
 from os import environ
+from collections import defaultdict
+from packaging.version import Version
 from rdflib_endpoint import SparqlEndpoint
 from rdflib import Graph, ConjunctiveGraph, URIRef, Literal
 from fastapi import FastAPI, Request, Response
@@ -12,11 +15,12 @@ from fastapi.responses import JSONResponse
 from fastapi.staticfiles import StaticFiles
 from codemeta import __path__ as CODEMETAPATH
 from codemeta.codemeta import serialize
-from codemeta.common import getstream, init_graph, AttribDict, SDO, RDF, CODEMETAPY
+from codemeta.validation import get_validation_report
+from codemeta.common import getstream, init_graph, AttribDict, SDO, RDF, CODEMETAPY, urijoin
 from codemeta.parsers.jsonld import parse_jsonld
 import argparse
 
-VERSION = "0.1"
+VERSION = "0.3.0" #also adapt in setup.py
 
 STATIC_DIR = os.path.join(CODEMETAPATH[0], "resources")
 
@@ -31,6 +35,8 @@ PREFIX spdx: <http://spdx.org/licences/>
 PREFIX orcid: <http://orcid.org/>
 """
 
+    
+
 class CodemetaServer(FastAPI):
     def __init__(self, *args,
                  graph: str,
@@ -39,6 +45,8 @@ class CodemetaServer(FastAPI):
                  baseurl: str = "http://localhost:8080",
                  baseuri: str = "",
                  version: str = VERSION,
+                 includecontext: bool = False,
+                 addcontext: list = [],
                  **kwargs
                 ) -> None:
         """Constructor for the CodemetaServer"""
@@ -48,16 +56,29 @@ class CodemetaServer(FastAPI):
         else:
             self.baseuri = self.baseurl
 
+        if self.baseuri[-1] not in ('/','#','?'):
+            self.baseuri += "/"
+
         self.title = kwargs.get('title')
-        self.css = [ x.strip() for x in kwargs.get('css',"").split(",") if x.strip() ]
+        if kwargs.get('css'):
+            self.css = [ x.strip() for x in kwargs.get('css',"").split(",") if x.strip() ]
+        else:
+            self.css = []
+
+        self.intro = kwargs.get('intro',"")
 
         print(f"Instantiating codemeta server: graph={graph}, baseuri={self.baseuri}, baseurl={self.baseurl}",file=sys.stderr)
 
-        g, contextgraph = init_graph()
+        self.includecontext = includecontext
+        self.addcontext = addcontext
+        g, contextgraph = init_graph(self.get_args())
         parse_jsonld(g, None, getstream(graph), self.get_args())
         self.graph = g
         self.contextgraph = contextgraph
-        if 'inputlogdir' in kwargs:
+        if self.includecontext:
+            self.graph += contextgraph #include context
+        self.build_versionmap()
+        if kwargs.get('inputlogdir'):
             self.read_logs(kwargs['inputlogdir'])
         # Instantiate FastAPI
         super().__init__(
@@ -96,6 +117,22 @@ class CodemetaServer(FastAPI):
         async def index(request: Request, res: Optional[str] = None, q: Optional[str] = None, sparql: Optional[str] = None):
             return self.get_index(request,res,q,sparql, "cardindex.html", )
 
+
+        @self.get("/services/",
+                  name="Services",
+                  description="Provides a visualised index focused on software services",
+                  responses= {
+                      200: {
+                          "description": "Service index",
+                          "content": {
+                              "text/html": {},
+                          }
+                      },
+                  }
+                )
+        async def services(request: Request, res: Optional[str] = None, q: Optional[str] = None, sparql: Optional[str] = None):
+            return self.get_index(request,res,q,sparql, "serviceindex.html", )
+
         @self.get("/table/",
                   name="Index",
                   description="Provides a tabular index to the data or simply returns all the data, i.e. the entire knowledge graph, depending on the content negotiation.",
@@ -149,6 +186,26 @@ class CodemetaServer(FastAPI):
                                   serialize(self.graph, None, self.get_args("turtle"), contextgraph=self.contextgraph )
                                )
 
+        @self.get("/validation/{resource:path}",
+                  name="Validation report",
+                  description="Returns a specific validation log",
+                  responses= {
+                      200: {
+                          "description": "Visualisation validation log",
+                          "content": {
+                              "text/plain": {},
+                          }
+                      },
+                  }
+                 )
+        async def get_validation(resource: str, request: Request):
+            res = URIRef(urijoin(self.baseuri, "validation", resource))
+            if (res,None,None) in self.graph:
+                return self.respond( "text",
+                             get_validation_report(self.graph, res)
+                            )
+            else:
+                return self.respond404("text")
 
         @self.get("/{resource:path}",
                   name="Resource",
@@ -174,13 +231,22 @@ class CodemetaServer(FastAPI):
                 output_type = "turtle"
             else:
                 output_type = self.get_output_type(request)
-            res = URIRef(os.path.join(self.baseuri, resource))
+            res = URIRef(urijoin(self.baseuri, resource))
             if (res,None,None) in self.graph:
                 return self.respond( output_type,
                              serialize(self.graph, res, self.get_args(output_type), contextgraph=self.contextgraph, title=self.title )
                             )
             else:
-                return self.respond404(output_type)
+                #if the resource does not exist, a version qualifier may be missing, attempt to find the latest version
+                versions = self.versionmap.get(resource.strip("/"),[])
+                if versions:
+                    res = URIRef(urijoin(self.baseuri, resource,versions[0])) #first version is the latest one
+                    if (res,None,None) in self.graph:
+                        return self.respond( output_type,
+                                     serialize(self.graph, res, self.get_args(output_type), contextgraph=self.contextgraph, title=self.title )
+                                    )
+            return self.respond404(output_type)
+
 
     def get_index(self, request: Request, res: Optional[str] = None, q: Optional[str] = None, sparql: Optional[str] = None, indextemplate: str  = "cardindex.html"):
         output_type = self.get_output_type(request)
@@ -206,7 +272,10 @@ class CodemetaServer(FastAPI):
             "output": output_type,
             "toolstore": True,
             "no_cache": True,
-            "css": [ os.path.join(self.baseurl,"static/codemeta.css") , os.path.join(self.baseurl,"static/fontawesome.css") ] + self.css
+            "includecontext": self.includecontext,
+            "addcontext": self.addcontext,
+            "intro": self.intro,
+            "css": [ urijoin(self.baseurl,"static/codemeta.css") , urijoin(self.baseurl,"static/fontawesome.css") ] + self.css
         })
 
     def respond(self, output_type: str, content: Union[str,bytes, None]) -> Response:
@@ -217,6 +286,7 @@ class CodemetaServer(FastAPI):
             return Response( content=content, media_type="text/turtle")
         elif output_type == "html":
             return Response( content=content, media_type="text/html")
+        return Response( content=content, media_type="text/plain")
 
     def respond404(self, output_type: str) -> Response:
         if output_type == 'json':
@@ -225,6 +295,7 @@ class CodemetaServer(FastAPI):
             return Response(status_code=404, content="", media_type="text/turtle")
         elif output_type == "html":
             return Response(status_code=404, content="<html><body><strong>404</strong> - Not Found - Resource does not exist</body></html>", media_type="text/html")
+        return Response(status_code=404, content="404 Not Found - Resource does not exist", media_type="text/plain")
 
     def respond400(self, output_type: str, message: str) -> Response:
         if output_type == 'json':
@@ -233,6 +304,7 @@ class CodemetaServer(FastAPI):
             return Response(status_code=400, content="", media_type="text/turtle")
         elif output_type == "html":
             return Response(status_code=400, content=f"<html><body><strong>400</strong> - Bad Request - {message}</body></html>", media_type="text/html")
+        return Response(status_code=400, content=f"400 Bad Request - {message}", media_type="text/plain")
 
     def get_output_type(self, request: Request) -> str:
         """Get the outputtype based on content negotiation"""
@@ -307,19 +379,50 @@ class CodemetaServer(FastAPI):
         for logfile in glob.glob(os.path.join(inputlogdir,"*.harvest.log")):
             identifier = os.path.basename(logfile).split(".")[0]
             res = URIRef(self.baseuri + identifier)
-            if (res,RDF.type,SDO.SoftwareSourceCode) in self.graph:
-                errors = 0
-                with open(logfile,'r',encoding='utf-8') as f:
-                    logdata = f.readlines()
-                    for line in logdata:
-                        if line.lower().find("harvester error") != -1:
-                            errors += 1
-                    self.graph.set((res, CODEMETAPY.errors, Literal(errors)))
-                    self.graph.set((res, CODEMETAPY.log, Literal("\n".join(logdata))))
-            else:
-                print(f"Log {logfile} describes non-existing resource {res}",file=sys.stderr)
+            if not (res,RDF.type,SDO.SoftwareSourceCode) in self.graph:
+                versions = self.versionmap.get(identifier,[])
+                if versions:
+                    res = URIRef(urijoin(self.baseuri, identifier,versions[0])) #first version is the latest one
+                    if not (res,RDF.type,SDO.SoftwareSourceCode) in self.graph:
+                        print(f"Log {logfile} describes non-existing resource {res}",file=sys.stderr)
+                        return
+            errors = 0
+            with open(logfile,'r',encoding='utf-8') as f:
+                logdata = f.readlines()
+                for line in logdata:
+                    if line.lower().find("harvester error") != -1:
+                        errors += 1
+                self.graph.set((res, CODEMETAPY.errors, Literal(errors)))
+                self.graph.set((res, CODEMETAPY.log, Literal("\n".join(logdata))))
 
+    def build_versionmap(self):
+        self.versionmap = defaultdict(list)
+        for s,_,_ in self.graph.triples((None,RDF.type,SDO.SoftwareSourceCode)):
+            if str(s).startswith(self.baseuri):
+                components = str(s)[len(self.baseuri):].strip("/").split("/")
+                if len(components) == 2:
+                    name, version = components
+                    print(f"Versionmap: adding SofwareSourceCode {name} with version {version}",file=sys.stderr)
+                    self.versionmap[name].append(version)
+        targetproducts = set()
+        for _,_, o in self.graph.triples((None,SDO.targetProduct,None)):
+            if str(o).startswith(self.baseuri) and o not in targetproducts:
+                targetproducts.add(o) #prevents duplicates
+                components = str(o)[len(self.baseuri):].strip("/").split("/")
+                if len(components) == 3:
+                    interfacetype, name, version = components
+                    print(f"Versionmap: adding {interfacetype} {name} with version {version}",file=sys.stderr)
+                    self.versionmap[interfacetype + "/" + name].append(version)
+        
+        for key, versions in self.versionmap.items():
+            self.versionmap[key] = sorted(versions, key=lambda x: Version(re.sub("^v","", x))._key if validversion(x) else (9999,0,0,0,0,0), reverse=True)
 
+def validversion(s: str) -> bool:
+    try:
+        Version(s)
+        return True
+    except:
+        return False
 
 def get_app(**kwargs):
     if not kwargs.get('graph'):
@@ -335,6 +438,12 @@ def get_app(**kwargs):
     if not kwargs.get('title'):
         if 'CODEMETA_TITLE' in environ:
             kwargs['title'] = environ['CODEMETA_TITLE']
+    if not kwargs.get('title'):
+        kwargs['title'] = "Codemeta server" #we must have a title
+
+    if not kwargs.get('intro'):
+        if 'CODEMETA_INTRO' in environ:
+            kwargs['intro'] = environ['CODEMETA_INTRO']
 
     if not kwargs.get('baseurl'):
         kwargs['baseurl'] = kwargs['baseuri']
@@ -347,6 +456,14 @@ def get_app(**kwargs):
         if 'CODEMETA_CSS' in environ:
             kwargs['css'] = environ['CODEMETA_CSS']
 
+    if not kwargs.get('addcontext'):
+        if 'CODEMETA_ADDCONTEXT' in environ:
+            kwargs['addcontext'] = environ['CODEMETA_ADDCONTEXT'].split(" ")
+
+    if not kwargs.get('includecontext'):
+        if 'CODEMETA_INCLUDECONTEXT' in environ:
+            kwargs['includecontext'] = environ['CODEMETA_INCLUDECONTEXT'].lower() in ("true","yes","1")
+
     return CodemetaServer(**kwargs)
 
 
@@ -359,6 +476,9 @@ def main():
     parser.add_argument('--baseuri',type=str, help="Base URI used in the IDs of all resources", action='store', required=True)
     parser.add_argument('--baseurl',type=str, help="Base URL (only needed when distinct from base URI)", action='store')
     parser.add_argument('--inputlogdir',type=str, help="Directory where *.harvest.log can be found as generated by codemeta-harvester", action='store')
+    parser.add_argument('--addcontext', help="Add the specified jsonld (must be a URL) to the context (and to the context graph). May be specified multiple times.", action='append',required=False)
+    parser.add_argument('--includecontext', help="Include all context vocabularies in the main graph and express it verbosely in serialisations. This makes the resoluting codemeta.json richer without the need to query certain external vocabularies, at the cost of added redundancy.", action='store_true',required=False)
+    parser.add_argument('--intro', type=str, help="Introductory text (html) to add to indices", action='store',required=False)
     parser.add_argument('--title',type=str, help="Title", action='store')
     parser.add_argument('--css',type=str, help="URLs to extra CSS stylesheets to use (comma separated list)", action='store')
     args = parser.parse_args() #parsed arguments can be accessed as attributes
